From 06689ada5011e1abeadaa0d33538e0b4069e5723 Mon Sep 17 00:00:00 2001
From: Melanie Plageman <melanieplageman@gmail.com>
Date: Thu, 20 Mar 2025 16:09:41 -0400
Subject: [PATCH v10 2/2] test backtrack

---
 src/backend/access/nbtree/nbtree.c            |  4 +
 src/test/modules/test_misc/meson.build        |  1 +
 .../modules/test_misc/t/008_vacuum_btree.pl   | 96 +++++++++++++++++++
 3 files changed, 101 insertions(+)
 create mode 100644 src/test/modules/test_misc/t/008_vacuum_btree.pl

diff --git a/src/backend/access/nbtree/nbtree.c b/src/backend/access/nbtree/nbtree.c
index 353600414a2..be29c4e70e2 100644
--- a/src/backend/access/nbtree/nbtree.c
+++ b/src/backend/access/nbtree/nbtree.c
@@ -32,6 +32,7 @@
 #include "storage/lmgr.h"
 #include "utils/fmgrprotos.h"
 #include "utils/index_selfuncs.h"
+#include "utils/injection_point.h"
 #include "utils/memutils.h"
 
 
@@ -1178,6 +1179,7 @@ backtrack:
 	attempt_pagedel = false;
 	backtrack_to = P_NONE;
 
+	INJECTION_POINT("nbtree-vacuum-page");
 	_bt_lockbuf(rel, buf, BT_READ);
 	page = BufferGetPage(buf);
 	opaque = NULL;
@@ -1468,6 +1470,8 @@ backtrack:
 		/* check for vacuum delay while not holding any buffer lock */
 		vacuum_delay_point(false);
 
+		INJECTION_POINT("nbtree-vacuum-page-backtrack");
+
 		/*
 		 * We can't use _bt_getbuf() here because it always applies
 		 * _bt_checkpage(), which will barf on an all-zero page. We want to
diff --git a/src/test/modules/test_misc/meson.build b/src/test/modules/test_misc/meson.build
index 9c50de7efb0..ed0822f19ee 100644
--- a/src/test/modules/test_misc/meson.build
+++ b/src/test/modules/test_misc/meson.build
@@ -16,6 +16,7 @@ tests += {
       't/005_timeouts.pl',
       't/006_signal_autovacuum.pl',
       't/007_catcache_inval.pl',
+      't/008_vacuum_btree.pl',
     ],
   },
 }
diff --git a/src/test/modules/test_misc/t/008_vacuum_btree.pl b/src/test/modules/test_misc/t/008_vacuum_btree.pl
new file mode 100644
index 00000000000..d393e1809ce
--- /dev/null
+++ b/src/test/modules/test_misc/t/008_vacuum_btree.pl
@@ -0,0 +1,96 @@
+# Copyright (c) 2024, PostgreSQL Global Development Group
+
+# Test btree vacuum
+
+use strict;
+use warnings;
+use PostgreSQL::Test::Cluster;
+use Test::More;
+
+if ($ENV{enable_injection_points} ne 'yes')
+{
+	plan skip_all => 'Injection points not supported by this build';
+}
+
+my $node = PostgreSQL::Test::Cluster->new('node');
+$node->init;
+
+# Turn autovacuum off for the cluster. This could be done at the table level
+# instead. However, since this file exercises vacuum, turn off autovacuum
+# globally. This also allows use of non-local injection points in vacuum code.
+$node->append_conf('postgresql.conf', 'autovacuum = off');
+$node->start;
+
+# Check if extension injection_points is available
+if (!$node->check_extension('injection_points'))
+{
+	plan skip_all => 'Extension injection_points not installed';
+}
+
+$node->safe_psql('postgres', 'CREATE EXTENSION injection_points;');
+
+my $psql_session = $node->background_psql('postgres');
+
+# Create a table with an index filled with dead index entries
+$node->safe_psql('postgres', q[
+		create table test_backtrack (col1 int);
+		insert into test_backtrack select generate_series(1,800);
+		create index on test_backtrack(col1);
+		delete from test_backtrack;
+	]
+);
+
+# Attach to two injection points. The first one will allow us to stop between
+# vacuuming each index page. The second is our validation that we did backtrack.
+$psql_session->query_safe(
+	qq[
+    SELECT injection_points_set_local();
+    SELECT injection_points_attach('nbtree-vacuum-page', 'wait');
+    SELECT injection_points_attach('nbtree-vacuum-page-backtrack', 'wait');
+]);
+
+# Start a vacuum of the table and index.
+$psql_session->query_until(
+	qr/starting_bg_psql/,
+		q(\echo starting_bg_psql
+		vacuum (index_cleanup on, parallel 0) test_backtrack;
+	));
+
+# The index vacuum should be waiting on our first injection point and is yet to
+# process any pages.
+$node->wait_for_event('client backend', 'nbtree-vacuum-page');
+
+# Wake up vacuum so it can process the first index leaf page.
+$node->safe_psql('postgres', "SELECT injection_points_wakeup('nbtree-vacuum-page');");
+
+# The first index leaf page is now vacuumed and vacuum should be waiting again
+# on the first injection point.
+$node->wait_for_event('client backend', 'nbtree-vacuum-page');
+
+# Insert data while vacuum is waiting to process the next leaf page. The
+# inserted data will force a page split in which some tuples from unprocessed
+# leaf pages will be moved to the first already vacuumed leaf page.
+$node->safe_psql('postgres',
+	"insert into test_backtrack select generate_series(1,300);");
+
+# Now we want the vacuum to continue. We don't want to wait on our first break
+# point again.
+# We need to make sure we are waiting before detaching and issuing a wakeup.
+# Otherwise there could be a race and the backend may not get woken up.
+$node->wait_for_event('client backend', 'nbtree-vacuum-page');
+$node->safe_psql('postgres', "SELECT injection_points_detach('nbtree-vacuum-page');");
+$node->safe_psql('postgres', "SELECT injection_points_wakeup('nbtree-vacuum-page');");
+
+# Wait on our second break point. Vacuum should have been forced to backtrack
+# and vacuum the first leaf page again to ensure it removed all dead index
+# entries.
+$node->wait_for_event('client backend', 'nbtree-vacuum-page-backtrack');
+
+# Once we wait on our second break point, we're done. Time to tell the backend
+# to detach and wake it up.
+$node->safe_psql('postgres', "SELECT injection_points_detach('nbtree-vacuum-page-backtrack');");
+$node->safe_psql('postgres', "SELECT injection_points_wakeup('nbtree-vacuum-page-backtrack');");
+
+ok($psql_session->quit);
+
+done_testing();
-- 
2.34.1

